==> algo-lang-1.1.3.tgz: package/lib/visitors/executor.js <==
import {EmptyVisitor} from './empty.js'
import {JungleError} from '../jungle/error.js'

class VariablesScope {
  constructor() {
    this.scopes = [{}]
  }

  get scope() {
    return this.scopes[this.scopes.length - 1]
  }

  pushScope() {
    this.scopes.push({})
  }

  popScope() {
    if (this.scopes.length === 1)
      throw new Error('Cannot pop last scope')

    this.scopes.pop()
  }

  add(name, type) {
    return this.scope[name] = {type: type, value: type.create()}
  }

  alias(name, expr) {
    return this.scope[name] = expr
  }

  get(name) {
    for (let i = this.scopes.length - 1; i >= 0; --i)
      if (name in this.scopes[i])
        return this.scopes[i][name]
    throw new Error('Variable ' + name + ' non trouvée')
  }
}

const binaryOps = {
  '←': function(left, right) {
    left = left()
    left.value = left.type.create(right().value)
  },
  '+': function(left, right, resType) {
    return resType.create(left().value + right().value)
  },
  '-': function(left, right, resType) {
    return resType.create(left().value - right().value)
  },
  '*': function(left, right, resType) {
    return resType.create(left().value * right().value)
  },
  '/': function(left, right, resType) {
    return resType.create(left().value / right().value)
  },
  'div': function(left, right, resType) {
    return resType.create(Math.floor(left().value / right().value))
  },
  'mod': function(left, right, resType) {
    return resType.create(left().value % right().value)
  },
  '=': function(left, right, resType) {
    return resType.create(left().value === right().value)
  },
  '>': function(left, right, resType) {
    return resType.create(left().value > right().value)
  },
  '>=': function(left, right, resType) {
    return resType.create(left().value >= right().value)
  },
  '<': function(left, right, resType) {
    return resType.create(left().value < right().value)
  },
  '<=': function(left, right, resType) {
    return resType.create(left().value <= right().value)
  },
  '<>': function(left, right, resType) {
    return resType.create(left().value !== right().value)
  },
  'et': function(left, right, resType) {
    return resType.create(left().value && right().value)
  },
  'ou': function(left, right, resType) {
    return resType.create(left().value || right().value)
  },
  'oue': function(left, right, resType) {
    return resType.create(left().value !== right().value)
  }
}

export class Executor extends EmptyVisitor {
  get routine() {
    return this.routines[this.routines.length - 1]
  }

  array(arr) {
    for (let i = 0; !this.routine.returnValue && i < arr.length; ++i)
      arr[i].accept(this)
  }

  visitAlgorithm(algo) {
    if (!algo.typed)
      throw new Error('You must type your ast before executing it')

    this.variables = new VariablesScope()
    this.routines = [algo]
    this.heap = {}

    algo.execute(this)

    if (!Object.keys(this.heap).length)
      return

    console.error('Fuites mémoires détectées:')
    for (let key in this.heap)
      console.error(' + ' +  key + ' a fuité un ' + this.heap[key].type.name)
  }

  visitVariables(variables) {
    variables.names.forEach(name => this.variables.add(name, variables.type))
  }

  visitIdentifierRef(identifier) {
    if (identifier.writable)
      return this.variables.get(identifier.ref)
    return {
      type: identifier.resType,
      get value() {
        return identifier.resType.create(identifier.value || identifier.ref)
      },
    }
  }

  visitValue(value) {
    return {
      type: value.resType,
      get value() {
        return value.resType.create(value.value)
      },
    }
  }

  visitRecordAccess(record) {
    const res = record.record.accept(this)

    return {
      type: record.resType,
      get value() {
        return res.value[record.property]
      },
      set value(val) {
        return res.value[record.property] = record.resType.create(val)
      },
    }
  }

  visitArrayAccess(array) {
    const res = array.array.accept(this)

    const positions = array.positions.map(pos => pos.accept(this).value)
    const lastPos = positions.pop() - 1
    const lastCell = positions.reduce((acc, dim, i) => {
      if (dim < 1 || dim > acc.length)
        throw new JungleError('Accès hors des limites d\'un tableau sur sa dimension ' +
          (i + 1) + ': ' + dim + ' > ' + acc.length)
      return acc[dim - 1]
    }, res.value)
    if (lastPos < 0 || lastPos >= lastCell.length)
      throw new JungleError('Accès hors des limites d\'un tableau sur sa dimension ' +
        (positions.length + 1) + ': ' + (lastPos + 1) + ' > ' + lastCell.length)

    return {
      type: array.resType,
      get value() {
        return lastCell[lastPos]
      },
      set value(val) {
        return lastCell[lastPos] = array.resType.create(val)
      },
    }
  }

  visitUnary(unary) {
    switch (unary.op) {
    case '+':
      return unary.expr.accept(this)
    case '-':
      const minus = unary.resType.create(-unary.expr.accept(this).value)
      return {
        type: unary.resType,
        value: minus,
      }
    case 'non':
      const not = unary.resType.create(!unary.expr.accept(this).value)
      return {
        type: unary.resType,
        value: not,
      }
    default:
      throw new Error('Implémentation manquante pour l\'opérateur unaire ' + unary.op)
    }
  }

  visitBinary(binary) {
    if (!(binary.op in binaryOps))
      throw new Error('Implémentation manquante pour l\'opérateur binaire ' + binary.op)

    const res = binaryOps[binary.op](() => binary.left.accept(this), () => binary.right.accept(this), binary.resType)

    return {
      type: binary.resType,
      get value() {
        return res;
      }
    }
  }

  visitPointerAccess(pointer) {
    const res = pointer.pointer.accept(this)

    if (res.value === 0)
      throw new JungleError('Null pointer exception')

    if (!(res.value in this.heap))
      throw new JungleError('Segmentation fault')

    const heap = this.heap

    return {
      type: pointer.resType,
      get value() {
        return heap[res.value].value
      },
      set value(val) {
        return heap[res.value].value = val
      },
    }
  }

  visitSi(si) {
    if (si.expr.accept(this).value)
      this.array(si.vrai)
    else if (si.faux)
      this.array(si.faux)
  }

  visitPour(pour) {
    const ref = this.variables.get(pour.ref)
    const start = pour.start.accept(this).value
    const end = pour.end.accept(this).value + (pour.desc ? -1 : 1)

    ref.value = start

    if (!pour.desc && start >= end || pour.desc && start <= end)
      return;

    const incr = pour.desc ? -1 : 1
    for (; !this.routine.returnValue && ref.value !== end; ref.value += incr) {
      this.array(pour.loop)
    }
  }

  visitTantQue(tantque) {
    let firstLoop = tantque.oneLoop

    while (!this.routine.returnValue && (firstLoop || tantque.expr.accept(this).value)) {
      firstLoop = false
      this.array(tantque.loop)
    }
  }

  visitRoutineCall(routine) {
    this.variables.pushScope()
    this.routines.push(routine.toCall)
    const res = this.routine.execute(this, routine.args.map(arg => arg.accept(this)))
    this.routines.pop()
    this.variables.popScope()
    return res
  }

  visitSelon(selon) {
    const testVal = selon.test.accept(this)

    for (var i = 0; !this.routine.returnValue && i < selon.expressions.length; ++i) {
      const expr = selon.expressions[i]
      const values = expr.accept(this)

      if (values.find(val => val.value === testVal.value)) {
        this.array(expr.expressions)
        break
      }
    }

    if (i === selon.expressions.length && selon.either)
      this.array(selon.either)
  }

  visitSelonExpression(expr) {
    return expr.scalars.map(scalar => scalar.accept(this))
  }

  visitScalar(scalar) {
    return scalar.value.accept(this)
  }
}

==> algo-lang-1.2.0.tgz: package/lib/visitors/executor.js <==
import {EmptyVisitor} from './empty.js'
import {JungleError} from '../jungle/error.js'

const binaryOps = {
  '←': function(left, right) {
    left = left()
    left.value = left.type.create(right().value)
  },
  '+': function(left, right, resType) {
    return resType.create(left().value + right().value)
  },
  '-': function(left, right, resType) {
    return resType.create(left().value - right().value)
  },
  '*': function(left, right, resType) {
    return resType.create(left().value * right().value)
  },
  '/': function(left, right, resType) {
    return resType.create(left().value / right().value)
  },
  'div': function(left, right, resType) {
    return resType.create(Math.floor(left().value / right().value))
  },
  'mod': function(left, right, resType) {
    return resType.create(left().value % right().value)
  },
  '=': function(left, right, resType) {
    return resType.create(left().value === right().value)
  },
  '>': function(left, right, resType) {
    return resType.create(left().value > right().value)
  },
  '>=': function(left, right, resType) {
    return resType.create(left().value >= right().value)
  },
  '<': function(left, right, resType) {
    return resType.create(left().value < right().value)
  },
  '<=': function(left, right, resType) {
    return resType.create(left().value <= right().value)
  },
  '<>': function(left, right, resType) {
    return resType.create(left().value !== right().value)
  },
  'et': function(left, right, resType) {
    return resType.create(left().value && right().value)
  },
  'ou': function(left, right, resType) {
    return resType.create(left().value || right().value)
  },
  'oue': function(left, right, resType) {
    return resType.create(left().value !== right().value)
  }
}

function checkArrayAccess(dim, pos, length) {
  if (1 <= pos && pos <= length)
    return
  throw new JungleError("Accès hors des limites d'un tableau sur sa dimension " + dim + ': ' + pos + (pos < 1 ? ' < 1 ' : ' > ' + length))
}

export class Executor extends EmptyVisitor {
  get routine() {
    return this.routines[this.routines.length - 1]
  }

  get variables() {
    return this.scopes[this.scopes.length - 1]
  }

  array(arr) {
    for (let i = 0; !this.routine.returnValue && i < arr.length; ++i)
      arr[i].accept(this)
  }

  visitAlgorithm(algo) {
    if (!algo.typed)
      throw new Error('You must type your ast before executing it')

    this.routines = [algo]
    this.scopes = []
    this.heap = {}

    algo.execute(this)

    if (!Object.keys(this.heap).length)
      return

    console.error('Fuites mémoires détectées:')
    for (let key in this.heap)
      console.error(' + ' +  key + ' a fuité un ' + this.heap[key].type.name)
  }

  visitVariables(variables) {
    variables.names.forEach(name => this.variables.add(name, variables.type))
  }

  visitIdentifierRef(identifier) {
    if (identifier.writable)
      return this.variables.get(identifier.ref)
    return {
      type: identifier.resType,
      get value() {
        return identifier.resType.create(identifier.value || identifier.ref)
      },
    }
  }

  visitValue(value) {
    return {
      type: value.resType,
      get value() {
        return value.resType.create(value.value)
      },
    }
  }

  visitRecordAccess(record) {
    const res = record.record.accept(this)

    return {
      type: record.resType,
      get value() {
        return res.value[record.property]
      },
      set value(val) {
        return res.value[record.property] = record.resType.create(val)
      },
    }
  }

  visitArrayAccess(array) {
    const res = array.array.accept(this)

    const positions = array.positions.map(pos => pos.accept(this).value)
    const lastPos = positions.pop() - 1
    const lastCell = positions.reduce((acc, pos, dim) => {
      checkArrayAccess(dim + 1, pos, acc.length);
      return acc[pos - 1]
    }, res.value)
    checkArrayAccess(positions.length + 1, lastPos + 1, lastCell.length)

    return {
      type: array.resType,
      get value() {
        return lastCell[lastPos]
      },
      set value(val) {
        return lastCell[lastPos] = array.resType.create(val)
      },
    }
  }

  visitUnary(unary) {
    switch (unary.op) {
    case '+':
      return unary.expr.accept(this)
    case '-':
      const minus = unary.resType.create(-unary.expr.accept(this).value)
      return {
        type: unary.resType,
        value: minus,
      }
    case 'non':
      const not = unary.resType.create(!unary.expr.accept(this).value)
      return {
        type: unary.resType,
        value: not,
      }
    default:
      throw new Error('Implémentation manquante pour l\'opérateur unaire ' + unary.op)
    }
  }

  visitBinary(binary) {
    if (!(binary.op in binaryOps))
      throw new Error('Implémentation manquante pour l\'opérateur binaire ' + binary.op)

    const res = binaryOps[binary.op](() => binary.left.accept(this), () => binary.right.accept(this), binary.resType)

    return {
      type: binary.resType,
      get value() {
        return res;
      }
    }
  }

  visitPointerAccess(pointer) {
    const res = pointer.pointer.accept(this)

    if (res.value === 0)
      throw new JungleError('Null pointer exception')

    if (!(res.value in this.heap))
      throw new JungleError('Segmentation fault')

    const heap = this.heap

    return {
      type: pointer.resType,
      get value() {
        return heap[res.value].value
      },
      set value(val) {
        return heap[res.value].value = val
      },
    }
  }

  visitSi(si) {
    if (si.expr.accept(this).value)
      this.array(si.vrai)
    else if (si.faux)
      this.array(si.faux)
  }

  visitPour(pour) {
    const ref = this.variables.get(pour.ref)
    const start = pour.start.accept(this).value
    const end = pour.end.accept(this).value + (pour.desc ? -1 : 1)

    ref.value = start

    if (!pour.desc && start >= end || pour.desc && start <= end)
      return;

    const incr = pour.desc ? -1 : 1
    for (; !this.routine.returnValue && ref.value !== end; ref.value += incr) {
      this.array(pour.loop)
    }
  }

  visitTantQue(tantque) {
    let firstLoop = tantque.oneLoop

    while (!this.routine.returnValue && (firstLoop || tantque.expr.accept(this).value)) {
      firstLoop = false
      this.array(tantque.loop)
    }
  }

  visitRoutineCall(routine) {
    const args = routine.args.map(arg => arg.accept(this))
    this.routines.push(routine.toCall)
    const res = this.routine.execute(this, args)
    this.routines.pop()
    return res
  }

  visitSelon(selon) {
    const testVal = selon.test.accept(this)

    for (var i = 0; !this.routine.returnValue && i < selon.expressions.length; ++i) {
      const expr = selon.expressions[i]
      const values = expr.accept(this)

      if (values.find(val => val.value === testVal.value)) {
        this.array(expr.expressions)
        break
      }
    }

    if (i === selon.expressions.length && selon.either)
      this.array(selon.either)
  }

  visitSelonExpression(expr) {
    return expr.scalars.map(scalar => scalar.accept(this))
  }

  visitScalar(scalar) {
    return scalar.value.accept(this)
  }
}
